/**
 * @class AttachmentUploader
 * @description Provides an interface for staging, unstaging, and uploading attachments.
 */
define([
    'dijit/Dialog',
    'dojo/aspect',
    'dijit/registry',
    'dojo/query',
    '../../LoadingSpinner/LoadingSpinner',
    '../EditCategoryDialog',
    '../DuplicateCategoryDialog',
    '../NewCategoryDialog',
    '../CustomDialog',
    'dojo/require',
    'dojo/NodeList-manipulate',
    'dojo/_base/lang',
    'fraudCommons/widgets/CommentManager/Comment/Attachment',
    'fraudCommons/widgets/CommentManager/CommentData',
    'dojox/form/uploader/FileList',
    'dojo/on',
    'dojo/has',
    'dojox/lang/functional',
    'dojo/dom-style',
    'dojo/i18n!./nls/validate',
    'dojox/form/Uploader',
    'dojo/dom-class',
    'dojo/dom-construct',
    'fraudCommons/util/NLSUtil',
    'dijit/form/Button',
    'dojo/text!./templates/AttachmentUploader.html',
    'dijit/_WidgetsInTemplateMixin',
    'dijit/_TemplatedMixin',
    'dijit/_WidgetBase',
    'dojox/html/entities',
    'dojo/_base/declare'
], function (Dialog, aspect, registry, query, LoadingSpinner, EditCategoryDialog, DuplicateCategoryDialog, NewCategoryDialog, CustomDialog, require, NodeListManipulate, lang, Attachment, CommentData, FileList, on, has, functional, domStyle, validate, Uploader, domClass, domConstruct, NLSUtil, Button, template, WidgetsInTemplateMixin, TemplatedMixin, WidgetBase, entities, declare) {
    var containsEmptyFile = function (array) {
        var containsEmpty = false;
        functional.forEach(array, function (file, key) {
            if (file.size === 0) {
                containsEmpty = true;
            }
        });
        return containsEmpty;
    };

    return declare([WidgetBase, TemplatedMixin, WidgetsInTemplateMixin], {
        templateString: template,
        widgetsInTemplate: true,

        /**
         * @constructs AttachmentUploader
         * @param {number} [commentId] - If we are editing a comment, this is the comment's ID.
         * @param {function} dataCallback - The callback to provide comment and attachment information when interacting
         *                                  with the server.
         * @param {function} onComplete - The callback to execute after server interaction.
         * @param {function} cancelCallback - The callback to execute when the user chooses to cancel editing the
         *                                    comment.
         * @param {function} emptyFileCallback - The callback to execute if the user attempts to upload an empty file.
         * @param {boolean} editForm - True if we are editing a comment.
         */
        constructor: function (args) {
            this.commentId = args.commentId;
            this.dataCallback = args.dataCallback;
            this.onComplete = args.onComplete;
            this.editForm = args.editForm;
            this.cancelCallback = args.cancelCallback;
            this.emptyFileCallback = args.emptyFileCallback;
            this.attachmentsToRemove = [];
            this.newCategoriesAllowed = args.newCategoriesAllowed;
        },

        postCreate: function () {
            this.inherited(arguments);
            this.editFormCancelButton.set('label', NLSUtil.resolve('btn_cancel', validate));
            this.submitButton.set('label', NLSUtil.resolve(this.editForm ? 'label_update' : 'label_post', validate));
            var that = this;
            var isIE = has("ie");
            var oldAttachmentsList = domConstruct.create("ul");
            var urlString = 'comments';

            if(this.newCategoriesAllowed) {
                domClass.add(this.categorySelect.dropDown.domNode, 'newCategoriesAllowed');
            }

            if(this.categories) {
                this.categorySelect.addOption(this.categories);
                this.categorySelect.getOptions().some(function(c) {
                    if(c.value === that.subject) {
                        that.categorySelect.focusChild({option:c});
                        return true;
                    } else {
                        return false;
                    }
                });
                if (this.newCategoriesAllowed) {
                    this.categorySelect.addOption({value: "__newCategory", label: NLSUtil.resolve('new_category', validate)});
                }
            }

            this.own(this.categorySelect.watch('value', function(name, oldValue, newValue) {
                if(newValue === "__newCategory" && this.get('focused')) {
                    var newCategoryDialog = new CustomDialog({
                        title: NLSUtil.resolve('new_category', validate),
                        body: new NewCategoryDialog({
                            existingCategories: that.categorySelect.get('options')
                        }),
                        confirmButtonInitiallyDisabled: true,
//                        refocusNode: that.newComment.domNode,
                        confirmCallback: function() {
                            var newCategoryName = this.body.categoryName.get('value').trim();
                            var options = that.categorySelect.get('options');
                            var categoryValid = !options.some(function(o) {
                                return o.value.trim().toLowerCase() === newCategoryName.toLowerCase();
                            });
                            if (categoryValid) {
                                if (that.newCategoriesAllowed) {
                                    var nc = options.pop(); //remove "new category" temporarily...
                                    that.categorySelect.addOption({value: newCategoryName, label: entities.encode(newCategoryName)});
                                    that.categorySelect.newestCategory = newCategoryName;
                                    options.sort(function (a, b) {
                                        return a.label.localeCompare(b.label);
                                    });
                                    nc.label = NLSUtil.resolve('rename_category', validate, entities.encode(newCategoryName));
                                    nc.value = "__editCategory";
                                    options.push(nc); //...and push it back onto the end
                                }
                                that.categorySelect.set('options', options);
                                that.categorySelect.set('value', newCategoryName);
                                if(that.categorySelect.getOptions("")) {
                                    that.categorySelect.removeOption("");
                                }
                                return true;
                            } else {
                                var d = new CustomDialog({
                                    showCancelButton: false,
                                    title: NLSUtil.resolve('new_category', validate),
                                    body: new DuplicateCategoryDialog({
                                        existingCategories: that.categorySelect.get('options')
                                    })
                                });
                                d.show();
                                return false;
                            }
                        },
                        cancelCallback: function() {
                            that.categorySelect.set('value', oldValue);
                        }
                    });
                    newCategoryDialog.show();
                } else if(newValue === "__editCategory" && this.get('focused')) {
                    var editCategoryDialog = new CustomDialog({
                        title: NLSUtil.resolve('edit_category', validate),
                        body: new EditCategoryDialog({
                            category: that.categorySelect.newestCategory,
                            existingCategories: that.categorySelect.get('options')
                        }),
//                        refocusNode: that.newComment.domNode,
                        confirmCallback: function() {
                            var newCategoryName = this.body.categoryName.get('value').trim();
                            var options = that.categorySelect.get('options');

                            var categoryValid = !options.some(function(o) {
                                var category = o.value.trim();
                                if(newCategoryName.toLowerCase() === that.categorySelect.newestCategory.toLowerCase()) {
                                    return false;
                                } else {
                                    return category.toLowerCase() === newCategoryName.toLowerCase();
                                }
                            });
                            if(categoryValid) {
                                var nc = options.pop(); //remove "edit category" temporarily...
                                options.forEach(function(o) {
                                    if(o.value === that.categorySelect.newestCategory) {
                                        o.value = newCategoryName;
                                        o.label = entities.encode(newCategoryName);
                                    }
                                });
                                options.sort(function(a, b) {
                                    return a.label.localeCompare(b.label);
                                });
                                nc.label = NLSUtil.resolve('rename_category', validate, entities.encode(newCategoryName));
                                options.push(nc); //...and push it back onto the end
                                that.categorySelect.newestCategory = newCategoryName;
                                that.categorySelect.set('options', options);
                                that.categorySelect.set('value', newCategoryName);
                                return true;
                            } else {
                                var d = new CustomDialog({
                                    showCancelButton: false,
                                    title: NLSUtil.resolve('edit_category', validate),
                                    body: new DuplicateCategoryDialog({
                                        existingCategories: that.categorySelect.get('options')
                                    })
                                });
                                d.show();
                                return false;
                            }
                        },
                        cancelCallback: function() {
                            that.categorySelect.set('value', oldValue);
                        }
                    });
                    editCategoryDialog.show();
                }
//                that.newCommentSubject.set('value', newValue);
//                that.subject = newValue;
//                lang.hitch(that, _toggleSubmitButton)(that.newComment.get('value'), that.subject);
            }));

            domClass.add(oldAttachmentsList, "commentattachments");

            /* Create attachments for editing a comment */
            if (this.editForm) {
                urlString += '?action=update';
                domStyle.set(this.editFormAttachmentsContainer, 'display', 'inline-block');
                functional.forEach(this.attachments, function (attachment, key) {
                    var li = new Attachment({
                        title: attachment.name,
                        href: 'ViewAttachmentServlet?a_id=' + attachment.id
                    });
                    domConstruct.place(li.domNode, that.editFormAttachmentsContainer, "last");
                });
            }

            /* Uploader widget to prompt the user for a file to upload*/
            this.uploader = new dojox.form.Uploader({
                label: "",
                iconClass: "sasAttachIcon",
//                deferredUploading: 1024,
                multiple: (!isIE || isIE > 9), //force ie9 to upload files individually
                uploadOnSelect: false,
                url: urlString
            });

            domClass.add(this.uploader.domNode, "commentbrowsebutton");

            /* Displays the list of files queued for upload.*/
            this.list = new FileList({
                uploader: this.uploader,
                headerIndex: NLSUtil.resolve('filelist_index', validate),
                headerType: NLSUtil.resolve('filelist_type', validate),
                headerFilename: NLSUtil.resolve('filelist_filename', validate),
                headerFilesize: NLSUtil.resolve('filelist_filesize', validate),
                'class': "attachmentList"
            });

            this.formAttachmentsPane.addChild(this.list, 0);
            if (this.editForm) {
                domStyle.set(this.editFormAttachmentsPane.domNode, 'display', 'block');
            }
            this.buttonPane.addChild(this.uploader);

            /* Start the necessary components */
            this.uploader.startup();
            this.list.startup();
            this.categorySelect.startup();

            /*
             * Handles the display of the comment's attachments when editing a comment.
             */
            if (this.editForm && this.attachments.length > 0) {
                if(this.attachments.length === 1) {
                    this.editFormAttachmentsRemoveButton.set('label', NLSUtil.resolve('btn_remove_singular_edit_existing', validate));
                } else {
                    this.editFormAttachmentsRemoveButton.set('label', NLSUtil.resolve('btn_remove_plural_edit_existing', validate));
                }
                domStyle.set(oldAttachmentsList, { display: 'block' });
                domStyle.set(this.formAttachmentsPane.domNode, { display: 'block' });
                domStyle.set(this.editFormAttachmentsPane.domNode, { display: 'block' });
                functional.forEach(this.attachments, function (attachment) {
                    var li = new Attachment({
                        title: attachment.name,
                        href: 'ViewAttachmentServlet?a_id=' + attachment.id
                    });
                    domConstruct.place(li.domNode, oldAttachmentsList, "last");
                });
            } else {
                domStyle.set(this.formAttachmentsPane.domNode, { display: 'none' });
                domStyle.set(this.editFormAttachmentsPane.domNode, { display: 'none' });
            }

            /*
             * Submits the form to create a new comment or edit an existing comment.
             */
            on(this.submitButton, 'click', function () {
                domStyle.set(that.removeButton.domNode, 'display', 'none');

                that.submitButton.set('disabled', true);
                //newestCategory is tracking newly added category for the drop down.
                //If the user cancels we want to remove it. However, in this case we are posting
                //so we do not want to remove this, so by handshake we set it to undefined.
                that.categorySelect.newestCategory = undefined;

                if (that.editForm && that.attachmentsToRemove.length > 0) {
                    var completedLoadingFlag = false,
                        targetNodeArray = query('.commentContainer'),
                        targetNode = registry.byNode(targetNodeArray.length >= 1 ? targetNodeArray[0] : this.domNode);

                    new LoadingSpinner().target(targetNode).until(function() {
                        return completedLoadingFlag;
                    }).startSpinners(function() {});
                    CommentData.deleteAttachments(that.attachmentsToRemove, that.commentId, lang.hitch(that, that.upload, that.dataCallback())).then(function() {
                        completedLoadingFlag = true;
                    });
                } else {
                    that.upload(that.dataCallback());
                }

                if (that.editForm) {
                    that.form.destroy();
                    if (that.cancelCallback) {
                        that.cancelCallback();
                    }
                }
            });

            /*
             * When editing a comment, stages the comment's existing attachments to be removed.
             * @todo Allow unstaging one file at a time.
             */
            on(this.editFormAttachmentsRemoveButton, 'click', function () {
                functional.forEach(that.attachments, function (attachment, key) {
                    that.attachmentsToRemove.push(attachment.id);
                });
                domConstruct.empty(that.editFormAttachmentsContainer);
                domStyle.set(that.editFormAttachmentsPane.domNode, 'display', 'none');
                domClass.add(that.formAttachmentsPane.domNode, 'noBorder');
            });


            /*
             * Removes new attachments from the uploader's stage.
             */
            on(this.removeButton, 'click', function () {
                that.uploader.set('disabled', false);
                domStyle.set(that.uploader.domNode, "display", "inline-block");
                domStyle.set(that.formAttachmentsPane.domNode, {  display: 'none' });
                domStyle.set(that.removeButton.domNode, "display", "none");
                that.uploader.reset();
            });

            /*
             * When the uploader's value changes (i.e. when attachments are added/removed), show/hide the stage(s)
             * for the attachments.
             */
            on(this.uploader, "change", function () {
                if (that.uploader.getFileList().length > 0) {
                    if(that.uploader.getFileList().length === 1) {
                        that.removeButton.set('label', NLSUtil.resolve(that.editForm ? 'btn_remove_singular_edit' : 'btn_remove_singular', validate));
                    } else {
                        that.removeButton.set('label', NLSUtil.resolve(that.editForm ? 'btn_remove_plural_edit' : 'btn_remove_plural', validate));
                    }
                    if (containsEmptyFile(that.uploader.getFileList()) && that.emptyFileCallback) {
                        that.emptyFileCallback();
                    } else {
                        domStyle.set(that.formAttachmentsPane.domNode, {  display: 'block' });
                        if (that.attachmentsToRemove.length > 0) {
                            domStyle.set(oldAttachmentsList, "display", "none");
                        }
                        domStyle.set(that.uploader.domNode, "display", "none");
                        domStyle.set(that.attachmentPane.domNode, "display", "block");
                        domStyle.set(that.removeButton.domNode, 'display', 'block');
                    }
                } else {
                    domStyle.set(that.uploader.domNode, "display", "inline-block");
                    domStyle.set(that.attachmentPane.domNode, "display", "none");
                    domStyle.set(that.removeButton.domNode, 'display', 'none');
                }
            });

            /*
             * Clear the attachments when the Cancel button is clicked.
             */
            on(this.editFormCancelButton, 'click', function () {
                that.form.destroy();
                that.attachmentsToRemove = [];
                if (that.cancelCallback) {
                    that.cancelCallback();
                }
            });
        },

        /**
         * @description Uploads comment and attachment data to the server, then executes the onComplete callback.
         * @param {Object} data - The data to send to the server.
         */
        upload: function (data) {
            var that = this;
            var errorHandlerSignal;
            var completeHandlerSignal;

            var completedLoadingFlag = false,
                targetNodeArray = query('.commentContainer');
            var targetNode = null;
            targetNodeArray.some(function(element, i, nodeArray) {
                if(element.offsetHeight != 0) {
                    targetNode = element;
                    return true;
                } else {
                    return false;
                }
            });
            targetNode = registry.byNode(targetNode);

            new LoadingSpinner().target(targetNode).until(function() {
                return completedLoadingFlag;
            }).startSpinners(function() {});

            var completionFunction = function (response) {
                if (errorHandlerSignal) {
                    errorHandlerSignal.remove();
                }

                if (completeHandlerSignal) {
                    completeHandlerSignal.remove();
                }

                if (that.onComplete) {
                    that.onComplete(response);
                }

                completedLoadingFlag = true;
            };

            /*
             * TODO this is a hack ... apparently to indicate success the server needs to return
             * a textarea containing JSON data you want to return...
             * we decided the lesser evil is to just assume any error means "we're done"
             */
            errorHandlerSignal = on(this.uploader, 'error', completionFunction);
            completeHandlerSignal = on(this.uploader, 'complete', completionFunction);

            /*
             * Checks the uploader to see if attachments need to be uploaded.
             */
            if (this.uploader.getFileList().length > 0) {
                if(has("ie") === 9 && data) {
                    for(var nm in data) {
                        var _node = domConstruct.create('input', {name:nm, type:"hidden"});
                        _node.value = data[nm];
                        that.uploader._inputs.push(_node);
                        }
                    }
                this.uploader.upload(data);
            } else if (this.editForm) {
                CommentData.update(data, that.onComplete).then(function() {
                    completedLoadingFlag = true;
                });
            } else {
                CommentData.save(data, that.onComplete).then(function() {
                    completedLoadingFlag = true;
                });
            }
        },

        /**
         * @private
         * @description IE does not automatically refresh when the Flash plugin uploads files.
         *              This function forces the component to refresh after an upload through IE.
         * @param {Object} commentData - Data for the comment to poll.
         * @param {Object} newComment - Data for the new comment.
         */
        _pollingFunction: function (commentData, newComment) {
            var that = this;
            var pollInterval = 500; //every 1/2 second
            var response = null;

            var completedLoadingFlag = false,
                targetNodeArray = query('.commentContainer'),
                targetNode = registry.byNode(targetNodeArray.length >= 1 ? targetNodeArray[0] : this.domNode);

            new LoadingSpinner().target(targetNode).until(function() {
                return completedLoadingFlag;
            }).startSpinners(function() {});

            CommentData.hasAttachments(commentData, newComment, function (data) {
                if (!data) {
                    setTimeout(function () {
                        that._pollingFunction(commentData, newComment);
                    }, pollInterval);
                } else {
                    that.onComplete();
                }
            }).then(function() {
                completedLoadingFlag = true;
            });
        },

        /**
         * @description Disables the submit button.
         * @param {boolean} disabled - True to disable the submit button.
         */
        disableButtons: function (disabled) {
            this.submitButton.set('disabled', disabled);
        }
    });
});